use std::{error::Error, net::TcpListener, time::Duration};

use arti_client::{config::BoolOrAuto, StreamPrefs, TorClient, TorClientConfig};
use futures::{AsyncReadExt, Stream, StreamExt};
use log::{error, info};
use tor_hsservice::config::OnionServiceConfigBuilder;
use tor_hsservice::handle_rend_requests;
use axum::{
    http::StatusCode,
    Json, Router,
    debug_handler
};
use axum::routing;
use serde::{Deserialize, Serialize};
use futures::io::AsyncWriteExt;

use crate::tor_axum;

#[debug_handler]
async fn test_connection_to_tor(Json(payload): Json<SimpleMessage>) -> (StatusCode, Json<SimpleMessage>)  {
    info!("Got interesting payload! {}", payload.msg);
    (StatusCode::OK, Json(SimpleMessage{ msg: "World is hell indeed!".to_string() }))
}

#[derive(Serialize, Deserialize)]
struct SimpleMessage {
    msg: String
}

pub async fn start() -> Result<(), Box<dyn Error>> {
    let tor_client = TorClient::create_bootstrapped(TorClientConfig::default()).await?;

    let (onion_service, rend_requests) = tor_client.launch_onion_service(
        OnionServiceConfigBuilder::default()
            .nickname("gamal-service".to_owned().try_into().unwrap())
            .build()?,
    )?;
    
    let stream_requests = handle_rend_requests(rend_requests);
    
    let app = Router::new().route("/", routing::post(test_connection_to_tor));
    
    println!("serving at: http://{}", onion_service.onion_name().unwrap());
    
    let onion_name = onion_service.onion_name().unwrap();

    tokio::spawn(async move {
        tokio::time::sleep(Duration::from_secs(40)).await;

        let mut stream_prefs = StreamPrefs::new();
        stream_prefs.connect_to_onion_services(BoolOrAuto::Explicit(true));

        match tor_client.connect_with_prefs((onion_name.to_string(), 80), &stream_prefs).await {
            Ok(mut stream) => {

                let host = onion_name.to_string();
                let path = "/";

                // Construct the HTTP GET request manually
                let request = format!(
                    "POST {} HTTP/1.1\r\nHost: {}\r\nContent-Type: application/json\r\nConnection: close\r\n\r\n{}",
                    path, host, serde_json::to_string(&SimpleMessage { msg: "is world hell?".to_string()}).unwrap()
                );

                if let Ok(a) = stream
                    .write_all(request.as_bytes())
                    .await {
                        let _ = stream.flush().await;
                    let mut buf = [0; 4096];
                    if let Ok(n) = stream.read(&mut buf).await {
                        info!("Got response!!! {}", String::from_utf8(buf[..n].to_vec()).unwrap());
                    } else {
                        error!("No response!!!");
                    }
                        // Flushing the stream is important; see below!
                } else {
                    error!("GO NO MONEY");
                }
            }
            Err(err) => error!("Error: {}", err)
        }
    });
    
    tor_axum::serve(stream_requests, app).await;

    Ok(())
}